//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef OBJ_HPP
#define OBJ_HPP
//------------------------------------------------------------------------------
#include <SFML/Graphics.hpp>
//------------------------------------------------------------------------------
#include <luabind/object.hpp>
//------------------------------------------------------------------------------
#include "config.hpp"
//------------------------------------------------------------------------------
#include <vector>
#include <map>
#include <string>
//------------------------------------------------------------------------------
#include "Utils.hpp"

#include "ClassAllocCounter.hpp"
#include "ProtectedPtr.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
#define ADD_GET_TYPE_FN(type) virtual int GetType() const {return type;}
//------------------------------------------------------------------------------
#define X_OBJ_LIST \
    X(PhLine, OBJ_TYPE_PHYSIC_LINE) \
    X(PhRect, OBJ_TYPE_PHYSIC_RECT) \
    X(PhTriangle, OBJ_TYPE_PHYSIC_TRIANGLE) \
    X(Group, OBJ_TYPE_DRAWABLE_GROUP) \
    X(ParticleSource, OBJ_TYPE_DRAWABLE_PARTICULE_SRC) \
    X(Shape, OBJ_TYPE_DRAWABLE_SHAPE) \
    X(Sprite, OBJ_TYPE_DRAWABLE_SPRITE) \
    X(Text, OBJ_TYPE_DRAWABLE_TEXT) \
    X(PostFX, OBJ_TYPE_DRAWABLE_POSTFX) \
    X(Screen, OBJ_TYPE_DRAWABLE_SCREEN) \
    X(PhCircle, OBJ_TYPE_PHYSIC_CIRCLE) \
    X(PhShape, OBJ_TYPE_PHYSIC_SHAPE) \
    X(ScreenRenderer, OBJ_TYPE_DRAWABLE_SCREEN_RENDERER) \
    X(TexturedArea, OBJ_TYPE_DRAWABLE_TEXTURED_AREA) \

enum
{
    #define X(typeName, typeId) typeId ,
    X_OBJ_LIST
    #undef X
};

// Declaration
#define X(typeName, typeId) \
        class typeName;
    X_OBJ_LIST
#undef X
//------------------------------------------------------------------------------
class Group;
class Screen;
//------------------------------------------------------------------------------
template<typename T>
T floatingModulo(const T &nb, const T &divider)
{
    int result = static_cast<int>( nb / divider );
    return (nb - T(result) * divider);
}
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** The base of all physic and drawable object. */
/// LUA_HEADER_COPY_OFF ///
//------------------------------------------------------------------------------
class Object : public ClassAllocCounter, public ProtectedClassBase
{
public:
    Object(Group *parent, const std::string &name);

    virtual ~Object();

    /// LUA_EXPORT_ON(Object, ProtectedPtr<Object>) ///
    /** Compare two objects (return true if these objects are in fact the same object). */
    bool operator ==(const Object& obj) {return this == &obj;}

    // Clone & Create
    /** Create a clone of the object and return an handler to the clone. */
    luabind::object Clone() const;

    /** Create a clone of the object, change the parent and return an handler to the clone. */
    luabind::object CloneToGroup(Group *newParent) const;

    // Position
    /** Access the object's position */
    //LUA// luaProperty(Vector2D, position, GetPosition, SetPosition);
    /** Get the object's position */
    Vector2D GetPosition() const;
    /** Set the object's position */
    virtual void SetPosition(const Vector2D &pos);

    /** Access the object's rotation */
    //LUA// luaProperty(float, rotation, GetRotation, SetRotation);
    /** Get the object's rotation */
    float GetRotation() const;
    /** Set the object's rotation */
    virtual void SetRotation(float angle);

    /** Access the object's x position */
    //LUA// luaProperty(float, x, GetPositionX, SetPositionX);
    /** Access the object's y position */
    //LUA// luaProperty(float, y, GetPositionY, SetPositionY);
    /** Get the object's x position */
    float GetPositionX() const {return GetPosition().x;}
    /** Get the object's y position */
    float GetPositionY() const {return GetPosition().y;}
    /** Set the object's x position */
    void SetPositionX(float pos) {SetPosition(Vector2D(pos, GetPositionY()));}
    /** Set the object's y position */
    void SetPositionY(float pos) {SetPosition(Vector2D(GetPositionX(), pos));}

    /** Access the object's global position (position from the top left corner of the screen) */
    //LUA// luaProperty(Vector2D, globalPosition, GetGlobalPosition, SetGlobalPosition);
    /** Get the object's global position (position from the top left corner of the screen) */
    Vector2D GetGlobalPosition() const;
    /** Set the object's global position (position from the top left corner of the screen) */
    void SetGlobalPosition(const Vector2D &pos);

    /** Access the object's global rotation (rotation of all parents and the object) */
    //LUA// luaProperty(Vector2D, rotation, GetGlobalRotation, SetGlobalRotation);
    /** Get the object's global rotation (rotation of all parents and the object) */
    float GetGlobalRotation() const;
    /** Set the object's global rotation (rotation of all parents and the object) */
    void SetGlobalRotation(float rotation);

    // Move
    /** Rotate the object */
    void Rotate(float angle);
    /** Move the object */
    void Move(const Vector2D &mvt);

    // Name
    /** Access the object's name */
    //LUA// luaProperty(std::string, name, GetName, SetName);
    /** Get the object's name */
    std::string GetName() const;
    /** Set the object's name */
    void SetName(const std::string &str);

    // Speed
    /** Access the object's speed */
    //LUA// luaProperty(float, speed, GetSpeed, SetSpeed);
    /** Set the object's speed */
    void SetSpeed(const Vector2D &speed);
    /** Get the object's speed */
    Vector2D GetSpeed() const;

    /** Access the object's angular velocity */
    //LUA// luaProperty(float, angularVelocity, GetAngularVelocity, SetAngularVelocity);
    /** Set the object's angular velocity */
    void SetAngularVelocity(float speed);
    /** Get the object's angular velocity */
    float GetAngularVelocity() const;

    // Type
    /** Get the object's type id */
    virtual int GetType() const = 0;

    // Unique Id
    /** Get an uniq identificator related to this object */
    long GetUniqueId() const;


    // Parent
    /** Change the parent of the object */
    virtual void ChangeParent(Group *parent) = 0;
    /** Destroy the object */
    virtual void Destroy();
    /** Get the object's parent (can be a Group or a Screen) */
    luabind::object GetParent();

    // User data
    /** Set user data */
    void SetUserData(const std::string &id, const luabind::object &data);
    /** Get user data */
    luabind::object GetUserData(const std::string &id);
    /** Return true if the user data exists, false otherwise */
    bool UserDataExists(const std::string &id);

    // Screen
    /** Get the screen (main parent) of this object */
    ProtectedPtr<Screen> GetParentScreen();
    /// LUA_EXPORT_OFF ///

    // Clone
    virtual Object *GetRawCloneToGroup(Group *newParent) const = 0;

    // No Lua
    // Register a clone of an object to a group (newParent == NULL mean to keep the last parent)
    virtual void RegisterClone(Group *newParent) = 0;

    // Screen
    /** Get the screen (main parent) of this object without protection pointer*/
    Screen *GetParentScreenRaw();

    /** Get the object's parent, uncasted (could be a Screen) and without protection pointer */
    Group *GetParentRaw();

    // Mvt
    virtual void ProceedMvt(float time);

    // Polaire
    void GetPolarPosition(float &dist, float &angle) const;
    void SetPolarPosition(float dist, float angle);

    // Some object need to be noticed when the parent screen is changed
    virtual void OnChangeScreen(Screen *backScreen, Screen *newScreen);

    // Global pos refresh
    /** Refresh the global rotated pos and propagate it to children objects */
    virtual void PropagateGlobalRotatedPos(const Vector2D &parentPos, float parentAngle);

protected:
    void SetParent(Group *parent);

    /** Refresh the global rotated pos and propagate it to children objects, starting in this object */
    void RefreshGlobalRotatedPos();

    /** Cast an unknown object for a lua object */
    luabind::object CastObject(Object *obj) const;

protected:
    Group *m_parent;
    std::string m_name;

    std::map<std::string, luabind::object> m_userData;

private:
    // Only accessible via Get*/Set*
    Vector2D m_speed;
    float m_angularVelocity;

    Vector2D m_pos;
    float m_rotation;

    // global rotated pos => the result pos (position with rotation applied)
    Vector2D m_globalRotatedPos;

    // global rotation => the rotation sum
    float m_globalRotation;

private:
    Object& operator=(const Object &obj);
};
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
#endif // OBJ_HPP
