//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "PhysicObject.hpp"

#include "Group.hpp"
#include "Screen.hpp"
#include "PhCircle.hpp"

#include "MyException.hpp"

#include <lua.hpp>
#include <luabind/luabind.hpp>

#include <iostream>
#include <map>
#include <cmath>
//------------------------------------------------------------------------------
namespace En
{
using namespace std;

LinePtVect::LinePtVect(const Vector2D &cPt, const Vector2D &cVect) : pt(cPt), vect(cVect) {}
LinePtVect::LinePtVect(const LinePtPt &line) : pt(line.pt1), vect(line.pt2 - line.pt1) {}
float LinePtVect::GetSize() const {return sqrt(GetSqrSum());};
float LinePtVect::GetSqrSum() const {return vect.x * vect.x + vect.y * vect.y;};

LinePtPt::LinePtPt(const Vector2D &cPt1, const Vector2D &cPt2) : pt1(cPt1), pt2(cPt2) {}
LinePtPt::LinePtPt(const LinePtVect &line) : pt1(line.pt), pt2(line.pt + line.vect) {}
float LinePtPt::GetSize() const {return sqrt(GetSqrSum());};
float LinePtPt::GetSqrSum() const {return LinePtVect(*this).GetSqrSum();};

//------------------------------------------------------------------------------
PhysicObject::PhysicObject(Group *parent, int phGroupId, const std::string &name) : Object(parent, name), m_group(phGroupId)
{
    m_parent = NULL;
    ChangeParent(parent);
}
//------------------------------------------------------------------------------
/*PhysicObject::PhysicObject(const PhysicObject& src) : Object(src)
{
    m_group = src.m_group;
    m_fnName = src.m_fnName;
    GetParentRaw()->Add(this);
    GetParentScreenRaw()->GetPhysicWorld()->Attach(this);
}*/
//------------------------------------------------------------------------------
void PhysicObject::RegisterClone(Group *newParent)
{
    // We change the parent if it's needed
    if(newParent != NULL)
    {
        m_parent = newParent;
    }

    // We register the object
    GetParentRaw()->Add(this);

    // We notice the object that the parent has changed => Will attach the objet to the pysic world
    OnChangeScreen(NULL, GetParentScreenRaw());
}
//------------------------------------------------------------------------------
int PhysicObject::GetPhysicGroup() const
{
    return m_group;
}
//------------------------------------------------------------------------------
PhysicObject::~PhysicObject()
{

}
//------------------------------------------------------------------------------
void PhysicObject::OnIntersect(PhysicObject *obj1, PhysicObject *obj2, const Vector2D &pos, const Vector2D &penetrationVector)
{
    const luabind::object &luaFunction = m_luaFunction;

    // Objects may be destroyed during the event
    ProtectedPtr<PhysicObject> ptr1 = obj1;
    ProtectedPtr<PhysicObject> ptr2 = obj2;

    if(! luaFunction.is_valid())
    {
        return;
    }

    try
    {
        luabind::call_function<void>(luaFunction, CastObject(obj1), CastObject(obj2), pos, penetrationVector);
    }
    catch(luabind::error &e)
    {
        cerr << "Failed to run lua function !" << endl;
        cerr << "LUA error (error) :n" << lua_tostring(Global::GetLuaState(), -1) << endl;
        cerr << "What: " << e.what() << endl;
        cerr << "File : '" << __FILE__ << "' at line " << __LINE__ << endl;
        cerr << "Objects : obj1 name = '";
        if(ptr1.IsValid())
        {
            cerr << obj1->GetName();
        }
        else
        {
            cerr << "Deleted";
        }

        cerr << "', obj2 name = '";
        if(ptr2.IsValid())
        {
            cerr << obj2->GetName();
        }
        else
        {
            cerr << "Deleted";
        }

        cerr << "'." << endl;
    }
    catch(luabind::cast_failed &e)
    {
        cerr << "Failed to run lua function !" << endl;
        cerr << "LUA error (cast_failed) :n" << lua_tostring(Global::GetLuaState(), -1) << endl;
        cerr << "What: " << e.what() << endl;
        cerr << "File : '" << __FILE__ << "' at line " << __LINE__ << endl;
        cerr << "Objects : obj1 name = '";
        if(ptr1.IsValid())
        {
            cerr << obj1->GetName();
        }
        else
        {
            cerr << "Deleted";
        }

        cerr << "', obj2 name = '";
        if(ptr2.IsValid())
        {
            cerr << obj2->GetName();
        }
        else
        {
            cerr << "Deleted";
        }

        cerr << "'." << endl;
    }
    catch(std::exception &e)
    {
        cerr << "Failed to run lua function (std::exception) !" << endl;
        cerr << "What: " << e.what() << endl;
        cerr << "File : '" << __FILE__ << "' at line " << __LINE__ << endl;
        cerr << "Objects : obj1 name = '";
        if(ptr1.IsValid())
        {
            cerr << obj1->GetName();
        }
        else
        {
            cerr << "Deleted";
        }

        cerr << "', obj2 name '";
        if(ptr2.IsValid())
        {
            cerr << obj2->GetName();
        }
        else
        {
            cerr << "Deleted";
        }

        cerr << "'." << endl;
    }
}
//------------------------------------------------------------------------------
void PhysicObject::SetEventFunction(const luabind::object &luaFunction)
{
    if(luaFunction.is_valid() && luabind::type(luaFunction) == LUA_TFUNCTION)
    {
        m_luaFunction = luaFunction;
    }
    else
    {
        m_luaFunction = luabind::object();
    }
}
//------------------------------------------------------------------------------
void PhysicObject::projectPointToAxis(const Vector2D &axis, minMaxStr &minMax, const Vector2D &point) const
{
    float val = axis.GetDotProduct(point);

    minMax.minVal = min(val, minMax.minVal);
    minMax.maxVal = max(val, minMax.maxVal);
}
//------------------------------------------------------------------------------
bool PhysicObject::IsIntersect(const PhysicObject &obj) const
{
    Vector2D a, b;
    return Intersect(obj, a, b);
}
//------------------------------------------------------------------------------
/** Return the list of objects that collide the given physic group and information about the collision. The returned value is a table of \\ref CollisionInfo. */
luabind::object PhysicObject::IsIntersectWithGroup(int phGroup)
{
    std::vector<CollisionInfo> tbl = GetParentScreenRaw()->GetPhysicWorld()->IsIntersectWithGroup(*this, phGroup);

    luabind::object table = luabind::newtable(Global::GetLuaState());
    int i = 1;
    for(std::vector<CollisionInfo>::iterator it = tbl.begin(); it != tbl.end(); it++)
    {
        table[i] = *it;
        i++;
    }

    return table;
}
//------------------------------------------------------------------------------
PhysicObject::minMaxStr PhysicObject::ProjectLineShapeToAxis(const std::vector<Vector2D> &ptList, const Vector2D& axis) const
{
    minMaxStr minMax;

    std::vector<Vector2D>::const_iterator ptIt = ptList.begin();
    minMax.minVal = minMax.maxVal = ptIt->GetDotProduct(axis);
    ptIt++;
    for(; ptIt != ptList.end(); ptIt++)
    {
        projectPointToAxis(axis, minMax, *ptIt);
    }

    return minMax;
}
//------------------------------------------------------------------------------
PhysicObject::minMaxStr PhysicObject::ProjectCircleToAxis(const Vector2D &center, float radius, const Vector2D& axis) const
{
    minMaxStr minMax;

    float centerProjection = center.GetDotProduct(axis);
    minMax.minVal = centerProjection - radius;
    minMax.maxVal = centerProjection + radius;

    return minMax;
}
//------------------------------------------------------------------------------
void PhysicObject::GetCollisionInfoFrom(const PhysicObject &obj1, const PhysicObject &obj2, const axisListType &axisList, Vector2D &point, Vector2D &penetrationVector) const
{
    // => Il y a collision !
    // Determination du vecteur de pénétration => Pour faire sortir (=obj2 ici) de obj1

    float minMoveSize = obj1.GetMaxInteractionRadius() * obj2.GetMaxInteractionRadius(); // Déplacement minimum
    Vector2D minMoveAxis;

    for(axisListType::const_iterator it = axisList.begin(); it != axisList.end(); it++)
    {
        float val = it->second.minMax1.minVal - it->second.minMax2.maxVal; // Sortie vers la gauche
        if(abs(val) < abs(minMoveSize))
        {
            minMoveSize = val;
            minMoveAxis = it->first;
        }

        val = it->second.minMax1.maxVal - it->second.minMax2.minVal; // Sortie vers la droite
        if(abs(val) < abs(minMoveSize))
        {
            minMoveSize = val;
            minMoveAxis = it->first;
        }
    }

    penetrationVector = minMoveSize * minMoveAxis;


    if(axisList.size() < 2)
    {
        // Pour les cercles, c'est normal.
        if(obj1.GetType() == OBJ_TYPE_PHYSIC_CIRCLE && obj2.GetType() == OBJ_TYPE_PHYSIC_CIRCLE)
        {
            const PhCircle &circle1 = static_cast<const PhCircle&>(obj1);
            const PhCircle &circle2 = static_cast<const PhCircle&>(obj2);

            point = circle1.GetGlobalPosition() + (circle2.GetGlobalPosition() - circle1.GetGlobalPosition()) * ((circle1.GetRadius() + circle2.GetRadius()) / circle1.GetRadius());
        }
        else
        {
            cerr << "Warning: PhysicObject::Intersect() : axisList.size() < 2 !\nName1 = " << obj1.GetName() << "\nName2 = " << obj2.GetName() << endl;

            // On Prend le milieu du segment séparent les deux objets (imprécis au possible)...
            point = (obj1.GetGlobalPosition() + obj2.GetGlobalPosition()) / 2;
        }
    }
    else
    {
        // Pt d'intersection
        // On a besoin de 2 axes non // pour trouver le pt
        axisListType::const_iterator it1 = axisList.begin();
            // Milieu de la "surface 1D" de pénétration
        float dist1 = (max(it1->second.minMax1.minVal, it1->second.minMax2.minVal) + min(it1->second.minMax1.maxVal, it1->second.minMax2.maxVal)) / 2;

        axisListType::const_iterator it2 = axisList.end(); // Les deux axis les plus séparé pour avoir la meilleur précision (vecteurs les moins collinéaires possible / séparé de l'angle max)
        it2--;
        float dist2 = (max(it2->second.minMax1.minVal, it2->second.minMax2.minVal) + min(it2->second.minMax1.maxVal, it2->second.minMax2.maxVal)) / 2;

            // Intersection des deux perpendiculaires
        point = GetIntersectionPoint(dist1 * it1->first, it1->first.GetDirectNormal(), dist2 * it2->first, it2->first.GetDirectNormal());
    }
}
//------------------------------------------------------------------------------
/** Return a vector that alwas have x > 0 and (y > 0 if x == 0) */
Vector2D PhysicObject::GetSameDirectionVector(Vector2D vect) const
{
     if(vect.x < 0)
     {
         vect = - vect; // x tjr positif pour eviter 2 vecteurs colinéaires mais opposés.
     }
     else if((vect.x == 0) && (vect.y < 0))
     {
         vect = - vect; // y tjr positif si x = 0 pour eviter 2 vecteurs colinéaires mais opposés (Vector2D(0, 1) et Vector2D(0, -1)).
     }

     return vect;
}
//------------------------------------------------------------------------------
void PhysicObject::GetAxisFromLineShape(const std::vector<Vector2D> &ptList, axisListType &axisList) const
{
    for(size_t i = 0; i < ptList.size(); i++)
    {
         Vector2D vect = (ptList[i] - ptList[(i + 1) % ptList.size()]).GetDirection().GetDirectNormal();
         vect = GetSameDirectionVector(vect);

         axisList.insert(std::pair<Vector2D, AxisVectorData>(vect, AxisVectorData()));
    }
}
//------------------------------------------------------------------------------
void PhysicObject::GetAxisFromDisk(const Vector2D &center, const std::vector<Vector2D> &objPtList, axisListType &axisList) const
{
    // => Axe entre le centre du cercle et le point le + proche dans shape 1
    // Cf: http://www.metanetsoftware.com/technique/tutorialA.html
    // FIXME : On peut faire mieux (enlever cet axe dans lertain cas (cf lien) !
    std::vector<Vector2D>::const_iterator itMin = objPtList.begin();
    float minSquareSum = (*itMin - center).GetSquareSum();
    for(std::vector<Vector2D>::const_iterator it = objPtList.begin() + 1; it != objPtList.end(); it++)
    {
        float sqrSum = (*it - center).GetSquareSum();
        if(sqrSum < minSquareSum)
        {
            minSquareSum = sqrSum;
            itMin = it;
        }
    }

    Vector2D vect = (*itMin - center).GetDirection(); // /!\ On prend la direction sans faire la normale
    vect = GetSameDirectionVector(vect);

    axisList.insert(std::pair<Vector2D, AxisVectorData>(vect, AxisVectorData()));
}
//------------------------------------------------------------------------------
bool PhysicObject::Intersect(const PhysicObject &obj, Vector2D &pos, Vector2D &penetrationVector) const
{
    // FIXME : Optimisation
    // -> Cache line list ?

    // Max interaction distance
    if(LinePtPt(GetGlobalPosition(), obj.GetGlobalPosition()).GetSize() > (GetMaxInteractionRadius() + obj.GetMaxInteractionRadius()))
    {
        return false;
    }

    // Test
    if(obj.GetType() != OBJ_TYPE_PHYSIC_CIRCLE && GetType() != OBJ_TYPE_PHYSIC_CIRCLE)
    {
        const PhysicObjectLine &obj1 = static_cast<const PhysicObjectLine&>(obj);
        const PhysicObjectLine &obj2 = static_cast<const PhysicObjectLine&>(*this);

        std::vector<Vector2D> list1 = obj1.GetAllPoints();
        std::vector<Vector2D> list2 = obj2.GetAllPoints();

        if(list1.empty() || list2.empty())
        {
            return false;
        }

        // Theoreme des axes séparateurs
            // Recherche des axes
        axisListType axisList;

        GetAxisFromLineShape(list1, axisList); // Shape 1
        GetAxisFromLineShape(list2, axisList); // Shape 2

        // Projection
        for(axisListType::iterator it = axisList.begin(); it != axisList.end(); it++)
        {
            // Projection de la figure 1
            it->second.minMax1 = ProjectLineShapeToAxis(list1, it->first);

            // Projection de la figure 2
            it->second.minMax2 = ProjectLineShapeToAxis(list2, it->first);

            // Verif si collision sur l'axe
            if((it->second.minMax2.maxVal <= it->second.minMax1.minVal) || (it->second.minMax2.minVal >= it->second.minMax1.maxVal)) // Les formes ne se chevauchent pas sur l'axe
            {
                return false;
            }
        }

        // => Il y a collision !
        // Determination du vecteur de pénétration => Pour faire sortir *this (=obj2 ici) de obj
        // /!\ minMax2 correspond à this
        GetCollisionInfoFrom(obj1, obj2, axisList, pos, penetrationVector);

        // Ok
        return true;
    }
    else // Circle + ?
    {
        const PhCircle *circle1;
        const PhysicObject *obj2;
        if(obj.GetType() == OBJ_TYPE_PHYSIC_CIRCLE)
        {
            circle1 = static_cast<const PhCircle*>(&obj);
            obj2 = this;
        }
        else
        {
            circle1 = static_cast<const PhCircle*>(this);
            obj2 = &obj;
        }

        if(obj2->GetType() == OBJ_TYPE_PHYSIC_CIRCLE) // Circle * 2
        {
            const PhCircle *circle2 = static_cast<const PhCircle*>(obj2);

            // Theoreme des axes séparateurs
                // Recherche des axes
            axisListType axisList;
            axisList.insert(std::pair<Vector2D, AxisVectorData>((circle1->GetGlobalPosition() - circle2->GetGlobalPosition()).GetDirection(), AxisVectorData()));


            // Projection
            for(axisListType::iterator it = axisList.begin(); it != axisList.end(); it++)
            {
                // Projection de la figure 1
                it->second.minMax1 = ProjectCircleToAxis(circle1->GetGlobalPosition(), circle1->GetRadius(), it->first);

                // Projection de la figure 2
                it->second.minMax2 = ProjectCircleToAxis(circle2->GetGlobalPosition(), circle2->GetRadius(), it->first);

                // Verif si collision sur l'axe
                if((it->second.minMax2.maxVal <= it->second.minMax1.minVal) || (it->second.minMax2.minVal >= it->second.minMax1.maxVal)) // Les formes ne se chevauchent pas sur l'axe
                {
                    return false;
                }
            }

            // => Il y a collision !
            // On veut faire sortir *this de obj, Ici on a de quoi faire sortir circle2 (= this) de circle1 (=obj) => Ok !!
            GetCollisionInfoFrom(*obj2, *circle1, axisList, pos, penetrationVector);

            // Ok
            return true;
        }
        else // Circle + shape
        {
            const PhysicObjectLine &objLine2 = static_cast<const PhysicObjectLine&>(*obj2);

            std::vector<Vector2D> list1 = objLine2.GetAllPoints();

            if(list1.empty())
            {
                return false;
            }

            // Theoreme des axes séparateurs
                // Recherche des axes
            axisListType axisList;

            GetAxisFromLineShape(list1, axisList); // Shape 1
            GetAxisFromDisk(circle1->GetGlobalPosition(), list1, axisList); // Cirlce 1


            // Projection
            for(axisListType::iterator it = axisList.begin(); it != axisList.end(); it++)
            {
                // Projection de la figure 1
                it->second.minMax1 = ProjectLineShapeToAxis(list1, it->first);

                // Projection de la figure 2
                it->second.minMax2 = ProjectCircleToAxis(circle1->GetGlobalPosition(), circle1->GetRadius(), it->first);

                // Verif si collision sur l'axe
                if((it->second.minMax2.maxVal <= it->second.minMax1.minVal) || (it->second.minMax2.minVal >= it->second.minMax1.maxVal)) // Les formes ne se chevauchent pas sur l'axe
                {
                    return false;
                }
            }

            // => Il y a collision !
            // /!\ minMax2 correspond à this
            GetCollisionInfoFrom(*obj2, *circle1, axisList, pos, penetrationVector);

            // On veut faire sortir *this de obj, Ici on a de quoi faire sortir circle1 de obj2
            if(circle1 != this)
            {
                penetrationVector = -penetrationVector;
            }

            // Ok
            return true;
        }
    }


    return false;
}
//------------------------------------------------------------------------------
void PhysicObject::ChangeParent(Group *parent)
{
    // /!\ à copier pour PhysicObject quant modifié !!
    Screen *backScreen = NULL, *newScreen = NULL;

    if(m_parent != NULL)
    {
        backScreen = GetParentScreenRaw();
        m_parent->Remove(this);
    }

    // m_parent = parent; // If parent is NULL, the object wille be deleted and it may need to know its parent

    if(parent != NULL)
    {
        m_parent = parent;
        parent->Add(this);
        newScreen = GetParentScreenRaw();
    }

    OnChangeScreen(backScreen, newScreen);
}
//------------------------------------------------------------------------------
void PhysicObject::OnChangeScreen(Screen *backScreen, Screen *newScreen)
{
    if(backScreen != newScreen) // If we don't change screen, this is unusefull
    {
        if(backScreen != NULL)
        {
            backScreen->GetPhysicWorld()->Detach(this);
        }

        if(newScreen != NULL)
        {
            newScreen->GetPhysicWorld()->Attach(this);
        }
    }
}
//------------------------------------------------------------------------------
Vector2D PhysicObject::GetIntersectionPoint(const Vector2D &pos1, const Vector2D &direction1, const Vector2D &pos2, const Vector2D &direction2) const
{
    const float div = direction1.x * direction2.y - direction1.y * direction2.x;

    float ratio1 = (direction1.x * pos1.y - direction1.x * pos2.y - direction1.y * pos1.x + direction1.y * pos2.x) / div;

    return pos2 + direction2 * ratio1;
}
//------------------------------------------------------------------------------
bool PhysicObject::IsLineInter(const LinePtVect &l1, const LinePtVect &l2, Vector2D &pos) const
{
    // Cf http://www.siteduzero.com/forum-83-382906-p1-calcul-du-point-d-intersection-de-deux-segments.html

    const Vector2D &A = l1.pt;
    const Vector2D &C = l2.pt;

    const Vector2D &I = l1.vect;
    const Vector2D &J = l2.vect;

    const float div = I.x * J.y - I.y * J.x;
    if(div == 0) // Parallele !
    {
        return false;
    }
    else
    {
        float ratio1 = (I.x * A.y - I.x * C.y - I.y * A.x + I.y * C.x) / div;
        float ratio2 = (J.x * A.y - J.x * C.y - J.y * A.x + J.y * C.x) / div;

        if((0 <= ratio1 && ratio1 <= 1) && (0 <= ratio2 && ratio2 <= 1))
        {
            pos = A + I * ratio2;

            return true;
        }
        else
        {
            return false;
        }
    }
}
//------------------------------------------------------------------------------
luabind::object PhysicObject::GetEventFunction() const
{
    return m_luaFunction;
}
//------------------------------------------------------------------------------
/*bool PhysicObject::IsLineCircleInter(const LinePtVect &l1, const Vector2D &center, float radius, Vector2D &pos) const
{
    const Vector2D &A = l1.pt; // Line 1 : start point
    const Vector2D &I = l1.vect; // Line 1 : Vect

    const Vector2D &C = center; // Circle center

    float ratio = -(A.x*I.x - C.x*I.x + A.y*I.y - C.y*I.y) / (pow(I.x, 2) + pow(I.y, 2));

    // On the line
    ratio = ratio > 1.f ? 1.f : ratio;
    ratio = ratio < 0.f ? 0.f : ratio;

    float squareDist = pow(A.x - C.x + ratio * I.x, 2) + pow(A.y - C.y + ratio * I.y, 2);
    bool result = (squareDist <= pow(radius, 2));

    if(result)
    {
        pos = A + I * ratio;
    }

    return(result);
}*/
//------------------------------------------------------------------------------
void PhysicObjectLine::RenderShape(sf::RenderTarget &App) const
{
    vector<Vector2D> ptList = GetAllPoints();
    for(size_t j = 0; j < ptList.size(); j++)
    {
        LinePtPt line(ptList[j], ptList[(j + 1) % ptList.size()]);
        sf::Shape sfLine = sf::Shape::Line(line.pt1.x, line.pt1.y, line.pt2.x, line.pt2.y, 1, Color(0, 0, 255));
        App.Draw(sfLine);

        sf::Shape circle = sf::Shape::Circle(ptList[j].x, ptList[j].y, 4, Color(255, 255, 255));
        App.Draw(circle);
    }
}
//------------------------------------------------------------------------------
bool PhysicObjectLine::IsPointInside(const Vector2D &point)
{
    if(LinePtPt(GetGlobalPosition(), point).GetSize() > GetMaxInteractionRadius())
    {
        return false;
    }

    // Ray
    bool crossing = false; // even / pair
    Vector2D pos(0, 0);

    std::vector<Vector2D> list1 = GetAllPoints();

    Vector2D rayStart = GetGlobalPosition() + Vector2D(m_cachedMaxInterRadius, m_cachedMaxInterRadius);
    LinePtPt ray(rayStart, point);

    for(size_t i = 0; i < list1.size(); i++)
    {
        LinePtPt line(list1[i], list1[(i + 1) % list1.size()]);

        if(IsLineInter(line, ray, pos))
        {
            crossing = !crossing;
        }
    }

    return crossing;
}
//------------------------------------------------------------------------------
void PhysicObjectLine::RefreshMaxInterRadius()
{
    m_cachedMaxInterRadius = 0;

    Vector2D pos = GetGlobalPosition();
    std::vector<Vector2D> list = GetAllPoints();

    for(std::vector<Vector2D>::iterator it = list.begin(); it != list.end(); it++)
    {
        m_cachedMaxInterRadius = max(LinePtPt(pos, *it).GetSize(), m_cachedMaxInterRadius);
    }
}

//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
