//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef PROTECTED_PTR_HPP
#define PROTECTED_PTR_HPP
//------------------------------------------------------------------------------
#include "config.hpp"
#include "MyException.hpp"
//------------------------------------------------------------------------------
class ProtectedClassBase;
//------------------------------------------------------------------------------
struct ptrInfo
{
    ProtectedClassBase *ptr;
    int refCount;
    #if DEBUG == 1
    std::string objName;
    #endif
};
//------------------------------------------------------------------------------
template<class T>
class ProtectedPtr
{
public:
    // Constructors
    //------------------------------------------------------------------------------
    ProtectedPtr()
    {
        m_ptrInfo = NULL;
    }
    //------------------------------------------------------------------------------
    ProtectedPtr( const ProtectedPtr<T>& ptr )
    {
        m_ptrInfo = ptr.m_ptrInfo;
        if(m_ptrInfo != NULL)
        {
            m_ptrInfo->refCount++;
        }
    }
    //------------------------------------------------------------------------------
    ProtectedPtr<T> &operator=(const ProtectedPtr<T> &src)
    {
        // We delete this pointer / reduce the number of refCount
        if(m_ptrInfo != NULL)
        {
            if(m_ptrInfo->refCount == 1) // Only this pointer remains
            {
                delete m_ptrInfo;
            }
            else
            {
                m_ptrInfo->refCount--;
            }
        }

        // Now we make this pointer equal to the second
        m_ptrInfo = src.m_ptrInfo;
        if(m_ptrInfo != NULL)
        {
            m_ptrInfo->refCount++;
        }

        return *this;
    }
    //------------------------------------------------------------------------------
    ProtectedPtr( T* ptr )
    {
        if(ptr != NULL)
        {
            m_ptrInfo = ptr->GetPtrInfo();
            m_ptrInfo->refCount++;
        }
        else
        {
            m_ptrInfo = NULL;
        }
    }
    //------------------------------------------------------------------------------
    // Destructors
    virtual ~ProtectedPtr()
    {
        if(m_ptrInfo)
        {
            if(m_ptrInfo->refCount == 1) // Only this pointer remains
            {
                delete m_ptrInfo;
            }
            else
            {
                m_ptrInfo->refCount--;
            }
        }
    }
    //------------------------------------------------------------------------------
    // Get
    T* GetPtr() const
    {
        if( m_ptrInfo == NULL)
        {
            MY_THROW("You are trying to access a NULL Object !");
        }

        if( m_ptrInfo->ptr == NULL)
        {
            #if DEBUG == 1
            MY_THROW("You are trying to access a deleted Object !\nName = '" << m_ptrInfo->objName << "'.");
            #else
            MY_THROW("You are trying to access a deleted Object !");
            #endif
        }
        return static_cast<T*>(m_ptrInfo->ptr);
    }
    //------------------------------------------------------------------------------
    T* operator->() const
    {
        return GetPtr();
    }
    //------------------------------------------------------------------------------
    operator T*() const
    {
        return GetPtr();
    }
    //------------------------------------------------------------------------------
    // Valid
    bool IsValid() const
    {
        /*if( m_ptrInfo == NULL)
        {
            MY_THROW("You are trying to access a NULL Object !");
        }*/

        return (m_ptrInfo != NULL && m_ptrInfo->ptr != NULL);
    }
    //------------------------------------------------------------------------------
    bool IsNull() const
    {
        return (m_ptrInfo == NULL);
    }

protected:
    ptrInfo *m_ptrInfo;
};
//------------------------------------------------------------------------------
template<class T>
T* get_pointer( ProtectedPtr<T> ptr )
{
    return ptr.GetPtr();
}
//------------------------------------------------------------------------------
class ProtectedClassBase
{
public:
    ProtectedClassBase();
    ProtectedClassBase(const ProtectedClassBase &src);
    virtual ~ProtectedClassBase();

    ptrInfo *GetPtrInfo();
    bool HaveReferencedPtr() const;

private:
    ptrInfo *m_ptrInfo;
};
//------------------------------------------------------------------------------
#endif // PROTECTED_PTR_HPP
