//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
#include "Screen.hpp"
#include "Global.hpp"
#include <iostream>
//------------------------------------------------------------------------------
#include "MyException.hpp"
#include "nbStrConvert.hpp"
#include "Event.hpp"
#include "PseudoEvent.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
/** Create a basic config with a 1024*768 size, and set according ratio */
ScreenConfig::ScreenConfig() : isChanged(true), cameraBottomRight(Vector2D(1024, 768)), cameraTopLeft(Vector2D(0, 0)), screenSize(Vector2D(1024, 768)), isFixedRatio(true)
{
    forcedRatio = 1024.f / 768.f;
}
//------------------------------------------------------------------------------
/** Create a basic config with user size, and set according ratio */
ScreenConfig::ScreenConfig(const Vector2D &size) : isChanged(true), cameraBottomRight(size), cameraTopLeft(Vector2D(0, 0)), screenSize(size), isFixedRatio(true)
{
    forcedRatio = size.x / size.y;
}
//------------------------------------------------------------------------------
/** Set the screen size and the according ratio. Reset the camera */
void ScreenConfig::ChangeSize(const Vector2D &size)
{
    forcedRatio = size.x / size.y;
    screenSize = size;
    cameraTopLeft = Vector2D(0, 0);
    cameraBottomRight = size;

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Reset the camera to show the entire screen */
void ScreenConfig::CameraReset()
{
    cameraTopLeft = Vector2D(0, 0);
    cameraBottomRight = screenSize;

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Move the camera with the given offset */
void ScreenConfig::CameraMove(const Vector2D &mvt)
{
    cameraTopLeft += mvt;
    cameraBottomRight += mvt;

    isChanged = true;
}
//------------------------------------------------------------------------------

/** Zoom */
void ScreenConfig::CameraZoom(float addedZoom)
{
    CameraSetHalfSize(CameraGetHalfSize() / addedZoom);

    isChanged = true;
}
//------------------------------------------------------------------------------

/** Set zoom, using screen size */
void ScreenConfig::CameraSetZoom(float zoom)
{
    CameraReset();
    CameraZoom(zoom);

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Get camera zoom, using screen size as a reference */
Vector2D ScreenConfig::CameraGetZoom() const
{
    return Vector2D(GetScreenSize().x / CameraGetSize().x, GetScreenSize().y / CameraGetSize().y);
}
//------------------------------------------------------------------------------
void ScreenConfig::CameraZoom(Vector2D addedZoom)
{
    Vector2D halfSize = CameraGetHalfSize();
    CameraSetHalfSize(Vector2D(halfSize.x / addedZoom.x, halfSize.y / addedZoom.y));

    isChanged = true;
}
//------------------------------------------------------------------------------
void ScreenConfig::CameraSetZoom(Vector2D zoom)
{
    CameraReset();
    CameraZoom(zoom);

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Get zoom, using screen size */
/*float ScreenConfig::CameraGetZoom()
{

}*/
//------------------------------------------------------------------------------
/** Set the camera center */
void ScreenConfig::CameraSetCenter(const Vector2D &center)
{
    Vector2D cameraHalfSize = CameraGetHalfSize();

    cameraTopLeft = center - cameraHalfSize;
    cameraBottomRight = center + cameraHalfSize;

    isChanged = true;
}
//------------------------------------------------------------------------------

/** Set the camera half size */
void ScreenConfig::CameraSetHalfSize(const Vector2D &halfSize)
{
    Vector2D center = CameraGetCenter();
    cameraTopLeft = center - halfSize;
    cameraBottomRight = center + halfSize;

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Get the camera center */
Vector2D ScreenConfig::CameraGetCenter() const
{
    return Vector2D((cameraBottomRight.x + cameraTopLeft.x) / 2.f, (cameraBottomRight.y + cameraTopLeft.y) / 2.f);
}
//------------------------------------------------------------------------------

/** Get the camera half size */
Vector2D ScreenConfig::CameraGetHalfSize() const
{
    return CameraGetSize() / 2.f;
}
//------------------------------------------------------------------------------
/** Set the camera half size */
void ScreenConfig::CameraSetSize(const Vector2D &cameraSize)
{
    CameraSetHalfSize(cameraSize / 2.f);
}
//------------------------------------------------------------------------------
/** Get the camera half size */
Vector2D ScreenConfig::CameraGetSize() const
{
    return cameraBottomRight - cameraTopLeft;
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Screen::Screen(const std::string &name) : Group(NULL, name), m_isMouseRelative(false), m_backMousePos(0, 0), m_isCursorVisible(true), m_timeSpeed(1), m_backgroundColor(Color(0, 0, 0)), m_savedWindowSize(sf::Vector2i(-1, -1)), m_renderScreen(NULL)
{
    Global::AddScreen(this);
    RefreshStripView(Global::GetApp());
}
//------------------------------------------------------------------------------
Screen::Screen(const Screen &screen) : Group(screen), m_renderScreen(screen.m_renderScreen)
{
    // Va être recréé automatiquement lors de l'enregistrement des clones
    //PhysicWorld m_world;
    // On copie les sons
    std::list<sf::Sound> m_sndTbl; // Il faut une liste pour ne pas détruire les sons en cas de resize !

    // A copier
    for(vector<EventHandler*>::const_iterator it = screen.m_evtHandlerTbl.begin(); it != screen.m_evtHandlerTbl.end(); it++)
    {
        m_evtHandlerTbl.push_back((*it)->Clone());
    }

    for(vector<PseudoEventHandler*>::const_iterator it = screen.m_pseudoEvtHandlerTbl.begin(); it != screen.m_pseudoEvtHandlerTbl.end(); it++)
    {
        m_pseudoEvtHandlerTbl.push_back((*it)->Clone());
    }

    // Mouse
    m_isMouseRelative = screen.m_isMouseRelative;
    m_backMousePos = screen.m_backMousePos;
    m_isCursorVisible = screen.m_isCursorVisible;

    // Time Speed
    m_timeSpeed = screen.m_timeSpeed;

    // Background color
    m_backgroundColor = screen.m_backgroundColor;

    // View & strip & cfg
    m_screenView = screen.m_screenView;

    // Will be recreated
    // sf::Shape m_strip1, m_strip2;

    m_screenCfg = screen.m_screenCfg;
    m_savedWindowSize = screen.m_savedWindowSize;

    // Copy the elements of the Group
    for(DrawZPosMap::const_iterator it = screen.m_drawTbl.begin(); it != screen.m_drawTbl.end(); it++ )
    {
        for(DrawLowZPosMap::const_iterator it2 = it->second.begin(); it2 != it->second.end(); it2++ )
        {
            it2->second->GetRawCloneToGroup(this);
        }
    }

    // iterators are invalidated for vectors
    for(std::vector<PhysicObject*>::const_iterator it = screen.m_phTbl.begin(); it != screen.m_phTbl.end(); it++)
    {
        (*it)->GetRawCloneToGroup(this);
    }

    // Register the screen
    Global::AddScreen(this);
    RefreshStripView(Global::GetApp());
}
//------------------------------------------------------------------------------
void Screen::Destroy()
{
    Global::RemoveScreen(this);
    delete this;
}
//------------------------------------------------------------------------------
ProtectedPtr<EventHandlerBase> Screen::RegisterButtonEvent(const luabind::object &luaFunction, const std::string &keyName, bool pressed)
{
    EventHandler *evt = new ButtonEventHandler(luaFunction, keyName, pressed);
    m_evtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------

ProtectedPtr<EventHandlerBase> Screen::RegisterMouseMoveEvent(const luabind::object &luaFunction)
{
    EventHandler *evt = new MouseMoveEventHandler(luaFunction);
    m_evtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
ProtectedPtr<EventHandlerBase> Screen::RegisterButtonInZoneEvent(const luabind::object &luaFunction, MouseZoneEvent *zoneEvt, const std::string &keyName, bool pressed)
{
    ButtonInZoneEventHandler *evt = new ButtonInZoneEventHandler(luaFunction, zoneEvt, keyName, pressed);
    m_evtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
ProtectedPtr<Timer> Screen::RegisterTimer(const luabind::object &luaFunction, float interval, bool isRunning)
{
    Timer *evt = new Timer(luaFunction, interval, isRunning);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
ProtectedPtr<Chronometer> Screen::RegisterChronometer(const luabind::object &luaFunction, float totalTime, bool isRunning)
{
    Chronometer *evt = new Chronometer(luaFunction, totalTime, isRunning);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
ProtectedPtr<MouseZoneEvent> Screen::RegisterMouseZoneEvent(const luabind::object &luaFunction, PhysicObject *zone)
{
    MouseZoneEvent *evt = new MouseZoneEvent(luaFunction, zone);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
ProtectedPtr<EventHandlerBase> Screen::RegisterFrameEvent(const luabind::object &luaFunction)
{
    FrameEvent *evt = new FrameEvent(luaFunction);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
/** One call back event */
OneCallbackEvent *Screen::RegisterOneCallbackEvent(const luabind::object &luaFunction, std::vector<luabind::object> argList)
{
    OneCallbackEvent *evt = new OneCallbackEvent(luaFunction, argList);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
void Screen::DrawRatioStrip(sf::RenderTarget &App)
{
    App.Draw(m_strip1);
    App.Draw(m_strip2);
}
//------------------------------------------------------------------------------
void Screen::RefreshStripView(sf::RenderTarget &App)
{
    if(!m_screenCfg.IsFixedRatio())
    {
        sf::FloatRect rect(m_screenCfg.GetCameraTopLeft().x, m_screenCfg.GetCameraTopLeft().y, m_screenCfg.GetCameraBottomRight().x, m_screenCfg.GetCameraBottomRight().y);
        m_screenView.SetFromRect(rect);
    }
    else
    {
        // Hauteur ralative à la taille totale à enlever pour atteindre le goal ratio :
        float hDiffRatio = (float(App.GetHeight()) - float(App.GetWidth()) / m_screenCfg.GetForcedRatio()) / float(App.GetHeight());

        if(hDiffRatio >= 0) // il faut ajouter des bandes horizontales
        {
            sf::FloatRect rect(m_screenCfg.GetCameraTopLeft().x, m_screenCfg.GetCameraTopLeft().y, m_screenCfg.GetCameraBottomRight().x, m_screenCfg.GetCameraTopLeft().y + m_screenCfg.CameraGetSize().y / (1.f - hDiffRatio));
            float width = rect.GetWidth();
            float height = rect.GetHeight();

            //m_isVerticalStrip = false;
            float stripHeight = hDiffRatio * height / 2.f;
            m_strip1 = sf::Shape::Rectangle(Vector2D(0, 0), Vector2D(width, stripHeight), GetBackgroundColor(), 0);
            m_strip1.Move(m_screenCfg.GetCameraTopLeft());
            m_strip2 = sf::Shape::Rectangle(Vector2D(0, height - stripHeight), Vector2D(width, height), GetBackgroundColor(), 0);
            m_strip2.Move(m_screenCfg.GetCameraTopLeft());

            // Screen position
            SetPosition(Vector2D(0, stripHeight));

            // View
            m_screenView.SetFromRect(rect);
        }
        else // Bandes verticales
        {
            float vDiffRatio = (float(App.GetWidth()) - float(App.GetHeight()) * m_screenCfg.GetForcedRatio()) / float(App.GetWidth());

            sf::FloatRect rect(m_screenCfg.GetCameraTopLeft().x, m_screenCfg.GetCameraTopLeft().y, m_screenCfg.GetCameraTopLeft().x + m_screenCfg.CameraGetSize().x / (1.f - vDiffRatio), m_screenCfg.GetCameraBottomRight().y);
            float width = rect.GetWidth();
            float height = rect.GetHeight();

            //m_isVerticalStrip = true;
            float stripWidth = vDiffRatio * width / 2.f;
            m_strip1 = sf::Shape::Rectangle(Vector2D(0, 0), Vector2D(stripWidth, height), GetBackgroundColor(), 0);
            m_strip1.Move(m_screenCfg.GetCameraTopLeft());
            m_strip2 = sf::Shape::Rectangle(Vector2D(width - stripWidth, 0), Vector2D(width, height), GetBackgroundColor(), 0);
            m_strip2.Move(m_screenCfg.GetCameraTopLeft());

            // Screen position
            SetPosition(Vector2D(stripWidth, 0));

            // View
            m_screenView.SetFromRect(rect);
        }
    }

    m_screenCfg.isChanged = false;
    m_savedWindowSize = sf::Vector2i(App.GetWidth(), App.GetHeight());
}

//------------------------------------------------------------------------------
/*void Screen::RemoveEvent(EventHandler* event)
{
    std::vector<EventHandler*>::iterator it = find(m_evtHandlerTbl.begin(), m_evtHandlerTbl.end(), event);
    if(it == m_evtHandlerTbl.end())
    {
        MY_THROW("RemoveEvent: not found !");
    }

    cout << it - m_evtHandlerTbl.begin() << endl;

    delete (*it);
    m_evtHandlerTbl.erase(it);
}
//------------------------------------------------------------------------------
void Screen::RemoveEvent(PseudoEventHandler* event)
{
    std::vector<PseudoEventHandler*>::iterator it = find(m_pseudoEvtHandlerTbl.begin(), m_pseudoEvtHandlerTbl.end(), event);
    if(it == m_pseudoEvtHandlerTbl.end())
    {
        MY_THROW("RemoveEvent: not found !");
    }

    m_pseudoEvtHandlerTbl.erase(it);
}*/
//------------------------------------------------------------------------------
void Screen::AddPhysicGroup(int phGroupId1, int phGroupId2)
{
    m_world.AddPhysicGroup(phGroupId1, phGroupId2);
}
//------------------------------------------------------------------------------
Color Screen::GetBackgroundColor()
{
    return m_backgroundColor;
}
//------------------------------------------------------------------------------
void Screen::SetBackgroundColor(const Color &color)
{
    m_backgroundColor = color;
    m_savedWindowSize.x = -1; // Refresh shape
}
//------------------------------------------------------------------------------
/** Get the screen configuration */
ScreenConfig Screen::GetScreenConfig()
{
    return m_screenCfg;
}
//------------------------------------------------------------------------------
/** Set the screen configuration */
void Screen::SetScreenConfig(const ScreenConfig &screenConfig)
{
    m_screenCfg = screenConfig;
    m_screenCfg.isChanged = true; // Invalidate
}

//------------------------------------------------------------------------------
void Screen::SetTimeSpeed(float timeSpeed)
{
    m_timeSpeed = timeSpeed;
}
//------------------------------------------------------------------------------
float Screen::GetTimeSpeed() const
{
    return m_timeSpeed;
}
//------------------------------------------------------------------------------
void Screen::Run()
{
    sf::RenderWindow &App = Global::GetApp();

    // Physic
    ProceedMvt(App.GetFrameTime() * m_timeSpeed);
    GetPhysicWorld()->ProceedIntersection();

    // Event
    ProceedEvent(App);

    // Same thing for renderScreen
    if(!m_renderScreen.IsNull())
    {
        if(!m_renderScreen.IsValid())
        {
            MY_THROW("Screen::Run(): The given render screen was destroyed !\nIn Screen '" << GetName() << "'.")
        }

        // Physic
        m_renderScreen->ProceedMvt(App.GetFrameTime() * m_timeSpeed);
        m_renderScreen->GetPhysicWorld()->ProceedIntersection();

        // Event
        m_renderScreen->ProceedEvent(App);
    }
}
//------------------------------------------------------------------------------
void Screen::RenderNoRenderScreen(sf::RenderTarget &App)
{
    // Strips & view
    if(m_screenCfg.isChanged || m_savedWindowSize != sf::Vector2i(App.GetWidth(), App.GetHeight()))
    {
        RefreshStripView(App);
    }

    App.SetView(m_screenView);

    Group::Render(App);
    DrawRatioStrip(App);
}
//------------------------------------------------------------------------------
void Screen::Render(sf::RenderTarget &App)
{
    if(!m_renderScreen.IsNull())
    {
        if(!m_renderScreen.IsValid())
        {
            MY_THROW("Screen::Render(): The given render screen was destroyed !")
        }

        m_renderScreen->Render(App);

        // We need the good view for the events
        if(m_screenCfg.isChanged || m_savedWindowSize != sf::Vector2i(App.GetWidth(), App.GetHeight()))
        {
            RefreshStripView(App);
        }
        App.SetView(m_screenView);

    }
    else
    {
        RenderNoRenderScreen(App);
    }
}
//------------------------------------------------------------------------------
void Screen::ProceedEvent(sf::RenderWindow &App)
{
    // Pseudo event

    // FIXME : Liste à déléter à faire ! // Ou pas
    for(std::vector<PseudoEventHandler*>::iterator it = m_pseudoEvtHandlerTbl.begin(); it != m_pseudoEvtHandlerTbl.end(); it++)
    {
        if((*it)->ShouldBeDestroyed())
        {
            delete *it;
            it = m_pseudoEvtHandlerTbl.erase(it) - 1;
        }
        else
        {
            // We need to keep the pos in case event are created in the event function (iterators would be invalidated)
            size_t pos = it - m_pseudoEvtHandlerTbl.begin();
            (*it)->Match(this, Global::GetApp());
            it = m_pseudoEvtHandlerTbl.begin() + pos;
        }
    }

    // Event
    bool alreadyAMouseMoveEvt = false;

    sf::Event event;
    while(App.GetEvent(event))
    {
        // Only one mouse move event per frame
        if(event.Type == sf::Event::MouseMoved)
        {
            if(alreadyAMouseMoveEvt)
            {
                continue; // We skip this event
            }
            else
            {
                alreadyAMouseMoveEvt = true;
            }
        }

        // FIXME : Liste à déléter à faire !
        for(std::vector<EventHandler*>::iterator it = m_evtHandlerTbl.begin(); it != m_evtHandlerTbl.end(); it++)
        {
            if((*it)->ShouldBeDestroyed())
            {
                delete *it;
                it = m_evtHandlerTbl.erase(it) - 1;
            }
            else
            {
                // We need to keep the pos in case event are created in the event function (iterators would be invalidated)
                size_t pos = it - m_evtHandlerTbl.begin();
                (*it)->Match(event, this, Global::GetApp());
                it = m_evtHandlerTbl.begin() + pos;
            }
        }

        // Relache la souris si on perd le focus
        if(IsMouseRelative())
        {
            if(event.Type == sf::Event::GainedFocus)
            {
                SetMouseRelative(true, false);
            }
            else if(event.Type == sf::Event::LostFocus)
            {
                SetMouseRelative(false, false);
            }
        }

        if(event.Type == sf::Event::Closed)
        {
            Global::Quit();
        }
    }

    if(IsMouseRelative())
    {
        Global::GetApp().SetCursorPosition(Global::GetApp().GetWidth() / 2, Global::GetApp().GetHeight() / 2);
    }
}
//------------------------------------------------------------------------------
Screen::~Screen()
{
    // !! We need to destroy groups before events : some of them might try to destroy event when they are destroyed.
    // Disconect every object
    OnChangeScreen(this, NULL);

    // On doit supprimer le contenu du groupe avant que Screen ne devienne un simple groupe
    Clear();

    // Delete Event Handler
    for(std::vector<EventHandler*>::iterator it = m_evtHandlerTbl.begin(); it != m_evtHandlerTbl.end(); it++ )
    {
        delete *it;
    }
    m_evtHandlerTbl.clear();

    for(std::vector<PseudoEventHandler*>::iterator it = m_pseudoEvtHandlerTbl.begin(); it != m_pseudoEvtHandlerTbl.end(); it++ )
    {
        delete *it;
    }
    m_pseudoEvtHandlerTbl.clear();
}
//------------------------------------------------------------------------------
void Screen::OnGetFocus()
{
    ResumeAllSounds();

    // Mouse
    if(IsMouseRelative())
    {
        SetMouseRelative(true, false);
    }
    ShowCursor(m_isCursorVisible);
}
//------------------------------------------------------------------------------
void Screen::OnLoseFocus()
{
    PauseAllSounds();

    // Mouse
    if(IsMouseRelative())
    {
        SetMouseRelative(false, false);
    }

    // To do not have a deleted View
    sf::RenderWindow &App = Global::GetApp();
    App.SetView(App.GetDefaultView());
}
//------------------------------------------------------------------------------
void Screen::ChangeParent(Group *parent)
{
    MY_THROW("ChangeParent for Screen");
}
//------------------------------------------------------------------------------
void Screen::SetMouseRelative(bool relative, bool changeValue)
{
    if(IsCurrentRunningScreen())
    {
        if(!m_isMouseRelative && relative) // Change absolute to relative
        {
            m_backMousePos = sf::Vector2<int>(Global::GetApp().GetInput().GetMouseX(), Global::GetApp().GetInput().GetMouseY());
            Global::GetApp().SetCursorPosition(Global::GetApp().GetWidth() / 2, Global::GetApp().GetHeight() / 2); // Pr Eviter de heurter les bords.
        }
        else if(m_isMouseRelative && !relative) // Change relative to absolute
        {
            Global::GetApp().SetCursorPosition(m_backMousePos.x, m_backMousePos.y);
        }
    }

    if(changeValue)
    {
        m_isMouseRelative = relative;
    }
}
//------------------------------------------------------------------------------
bool Screen::IsMouseRelative() const
{
    return m_isMouseRelative;
}
//------------------------------------------------------------------------------
void Screen::ShowCursor(bool show)
{
    m_isCursorVisible = show;
    if(IsCurrentRunningScreen())
    {
        Global::GetApp().ShowMouseCursor(m_isCursorVisible);
    }
}
//------------------------------------------------------------------------------
bool Screen::IsCursorVisible() const
{
    return m_isCursorVisible;
}
//------------------------------------------------------------------------------
/** Get the number of event and pseudo event registered in this screen. */
int Screen::GetEventCount() const
{
    return m_evtHandlerTbl.size() + m_pseudoEvtHandlerTbl.size();
}
//------------------------------------------------------------------------------
/** Set the render screen and enable it.\\n
 * A render screen is a normal Screen that is just meant to be rendered while this screen is the current active screen.\\n
 * Usualy, a render screen contains multiple ScreenRenderer targeting this screen. It allows you to flip the screen.
 * \\param renderScreen the screen to be rendered or nil to render this screen */
void Screen::SetRenderScreen(luabind::object renderScreen)
{
    if(renderScreen.is_valid())
    {
        if(luabind::type(renderScreen) == LUA_TNIL)
        {
            m_renderScreen = NULL;
        }
        else
        {
            try
            {
                m_renderScreen = luabind::object_cast<ProtectedPtr<Screen> >(renderScreen);
            }
            catch(luabind::cast_failed &e)
            {
                MY_THROW("renderScreen : bad type (" << e.what() << ")");
            }
        }
    }
    else
    {
        MY_THROW("renderScreen : not valid object");
    }
}
//------------------------------------------------------------------------------
/** Return true if this Screen is the current running screen, false otherwise */
bool Screen::IsCurrentRunningScreen() const
{
    return Global::GetCurrentScreenRaw() == this;
}
//------------------------------------------------------------------------------
/** Get the render screen or nil if no render screen where set.\\n
 * See Screen::SetRenderScreen. */
luabind::object Screen::GetRenderScreen()
{
    if(m_renderScreen == NULL)
    {
        return luabind::object();
    }
    else
    {
        return luabind::object(Global::GetLuaState(), ProtectedPtr<Screen>(m_renderScreen));
    }
}
//------------------------------------------------------------------------------
/** Return the first physic group that is not used by any PhysicObject in this Screen */
int Screen::GetUnusedPhGroup() const
{
    return GetPhysicWorld()->GetUnusedPhGroup();
}
//------------------------------------------------------------------------------
ProtectedPtr<Screen> Screen::GetRenderScreenRaw()
{
    return m_renderScreen;
}
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
